/*
 * File:        BLEControllerCentral.ino
 * Description: BLE central device script for a DIY controller
 *              six push buttons:A, B, C, D, E, F
 *              a joy stick:X, Y, K(push button of joy stick)        
 * Date:        11/13/2023
 * Author:      Sayaka 
 * Hardware MCU: ESP32 DEVKIT V4 Board
 *              Gamepad:  HiLetgo Gamepad Joystick Shield
 *              LED for the BLE conection indicator
 */
 
#include <BLEDevice.h>

#define SPI_SPEED   115200          // SPI speed

/* BLE */
// See the following for generating UUIDs:
// https://www.uuidgenerator.net/
static BLEUUID serviceUUID("4fafc201-1fb5-459e-8fcc-c5c9c331914b");
static BLEUUID charUUID("beb5483e-36e1-4688-b7f5-ea07361b26a8");

BLERemoteCharacteristic* pRemoteCharacteristicTX;
BLEAdvertisedDevice* targetDevice; 
static boolean doConnect = false;
static boolean doScan = false;
bool deviceConnected = false;

/* Controller */
// Push button pins: a,b,c,d,e,f,k
const int buttonPin[7] = {27, 26, 25, 33, 16, 17, 32}; 
const char buttonName[7] = {'A', 'B', 'C', 'D', 'E', 'F', 'K'};
int buttonState[7];
int lastButtonState[7];
unsigned long lastDebounceTime[7];
unsigned long debounceDelay = 20; 

// Joystick pins: x,y
const int joyPin[2] = {35, 34};
int joyPos[2];
int lastjoyPos[2];
float joyWeight = 0.6;
int joyOffset[2] = {0,0};
int joyAveTimes = 50; // Number of times to get data for offset value
int joyThreshold = 30; // Threshold value to send data

// LED for the controller indicato
const int ledPin = 5;
unsigned long lastLedTime=0;
unsigned long ledOnTime = 200;
unsigned long ledOffTime = 800;

/* send data */
struct sendData {
    uint8_t abcdBtn;
    uint8_t efkBtn;
    uint16_t xyJoy[2];
};
struct sendData  data;

/* Declaration of Functions */
void doInitialize();
bool doPrepare();
int isPushbuttonClicked(int);

/******************** < Callback classes > ********************/
class ClientCallbacks: public BLEClientCallbacks {
    void onConnect(BLEClient* pClient) {
    };
    void onDisconnect(BLEClient* pClient) {
        deviceConnected = false;
    }
};


class AdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks {
    void onResult(BLEAdvertisedDevice advertisedDevice) {
        Serial.print("Advertised Device found: ");
        Serial.println(advertisedDevice.toString().c_str());
        if (advertisedDevice.haveServiceUUID()) {
            BLEUUID service = advertisedDevice.getServiceUUID();
            Serial.print("Service UUID: "); Serial.println(service.toString().c_str());
            if (service.equals(serviceUUID)) {
                BLEDevice::getScan()->stop();
                targetDevice = new BLEAdvertisedDevice(advertisedDevice);
                doConnect = true;
                doScan = true;
            }
        }
    }
};


/******************** < Mandatory functions > ********************/
void setup() {
    doInitialize();
    BLEDevice::init("");
    Serial.println("Client application start...");

    BLEScan* pBLEScan = BLEDevice::getScan();
    pBLEScan->setAdvertisedDeviceCallbacks(new AdvertisedDeviceCallbacks());
    // scan 10 sec
    pBLEScan->setActiveScan(true);
    pBLEScan->start(10);
    // scan after 3 sec rest until the device is found 
    while (doConnect == false){      
      delay(3000);
      //Serial.println("rescan start...");
      pBLEScan->start(10);
    }

}

void loop() {
    if (doConnect == true) {
        
        if (doPrepare()) {
            Serial.println("Connected to the BLE peripheral.");
            // first sending data is special
            data.abcdBtn = 0xFF;
            data.efkBtn = 0xFF;
            data.xyJoy[0] = joyOffset[0];
            data.xyJoy[1] = joyOffset[1];
            pRemoteCharacteristicTX->writeValue((uint8_t*)&data, sizeof(sendData));
            Serial.print(data.abcdBtn);
            Serial.print(",");
            Serial.print(data.efkBtn);
            Serial.print(",");
            Serial.print(data.xyJoy[0]);
            Serial.print(",");
            Serial.println(data.xyJoy[1]);
            digitalWrite(ledPin, HIGH);
        } else {
            Serial.println("Failed to connect BLE peripheral.");
        }
        doConnect = false;
    }

    if (deviceConnected) {
        bool f_send = false;
        data.abcdBtn = 0;
        data.efkBtn = 0;
        uint8_t buttonCheck = 0;
        for (int i = 0; i < 4; i++) {
          buttonCheck = (uint8_t) isPushbuttonClicked(i);
          data.abcdBtn += buttonCheck << (2 * (3-i));
          if (buttonCheck != 0){
            f_send = true;
          }
        }
        for (int i = 0; i < 3; i++) {
          buttonCheck = (uint8_t) isPushbuttonClicked(i+4);
          data.efkBtn += buttonCheck << (2 * (3-i));
          if (buttonCheck != 0){
            f_send = true;
          }
        }
        for (int i =0; i < 2; i++) {
           joyPos[i] = joyWeight * joyPos[i] +(1-joyWeight) * analogRead(joyPin[i]);
           data.xyJoy[i] = joyPos[i];
           data.xyJoy[i] = analogRead(joyPin[i]);
           if (abs(joyPos[i] - joyOffset[i]) > joyThreshold){
                f_send = true;
           }
        }
        // Send data only when the controller's status changes.
        if (f_send){
            pRemoteCharacteristicTX->writeValue((uint8_t*)&data, sizeof(sendData));
            Serial.print(data.abcdBtn);
            Serial.print(",");
            Serial.print(data.efkBtn);
            Serial.print(",");
            Serial.print(data.xyJoy[0]);
            Serial.print(",");
            Serial.println(data.xyJoy[1]);
        }
    } else if(doScan) {
        doConnect = true;
        digitalWrite(ledPin, LOW);
        BLEDevice::getScan()->start(0);

    }

    delay(10);
}

/******************** < Other functions > ********************/
/*  Initialization process  */
void doInitialize() {
    Serial.begin(SPI_SPEED);

    // set push buttons
    for (int i = 0; i < 7; i++) {
        buttonState[i] = HIGH;
        lastButtonState[i] = HIGH;
        lastDebounceTime[i] = 0;  
        pinMode(buttonPin[i], INPUT);
    }
    // set LED
    pinMode(ledPin, OUTPUT);
    digitalWrite(ledPin, LOW);

    Serial.println("BLE Client start ...");
}

/*  Processing to do only once when BLE connects */
bool doPrepare() {
    digitalWrite(ledPin, HIGH);
    
    BLEClient* pClient = BLEDevice::createClient();
    pClient->setClientCallbacks(new ClientCallbacks());
    Serial.println(" - Created client.");

    pClient->connect(targetDevice);
    Serial.println(" - Connected to peripheral.");

    BLERemoteService* pRemoteService = pClient->getService(serviceUUID);
    if (pRemoteService == nullptr) {
        Serial.print("Failed to find serviceUUID: ");
        Serial.println(serviceUUID.toString().c_str());
        pClient->disconnect();
        return false;
    }
    Serial.println(" - Found target service.");


    pRemoteCharacteristicTX = pRemoteService->getCharacteristic(charUUID);
    if (pRemoteCharacteristicTX == nullptr) {
      Serial.print("Failed to find charUUID: ");
      Serial.println(charUUID.toString().c_str());
      pClient->disconnect();
      return false;
    }
    Serial.println(" - Found characteristic charUUID.");

    // joystickの中央値を取得
    for (int i = 0; i < joyAveTimes; i++) {
        joyOffset[0] += analogRead(joyPin[0]);
        joyOffset[1] += analogRead(joyPin[1]);
        delay(50);
    }
    for (int i =0; i < 2; i++) {
        joyOffset[i] /= joyAveTimes;
        joyPos[i] = joyOffset[i];
    }

    Serial.print(joyOffset[0]);
    Serial.print(",");
    Serial.println(joyOffset[1]);

    deviceConnected = true;
    return true;
}

/* check the push button state */
// 0: no pushed
// 1: pushed
// 2: released 
int isPushbuttonClicked(int i) {

    int state = digitalRead(buttonPin[i]);
    if (state != lastButtonState[i]) { // reset timer
        lastDebounceTime[i] = millis();
    }
  
    if ((millis() - lastDebounceTime[i]) > debounceDelay) {
        if (state != buttonState[i]) {
            buttonState[i] = state;
            if (buttonState[i] == LOW) { // pushed
                return 1;
            }
            if (buttonState[i] == HIGH) { // released
                lastButtonState[i] = state;
                return 2;
            }
        }
    }
    lastButtonState[i] = state;
    return 0;
}
